use std::io::prelude::*;
use std::io::BufReader;
use std::process::{Command, Stdio};
use std::thread;

use xcb::render::Color;

use ui::context::Context;
use ui::color;
use widgets::{Message, MessageSender, Update, Widget, WidgetParams};

const MARGIN: u16 = 7;

struct Desktop {
    name: String,
    selected: bool,
    urgent: bool,
    position: u16,
    width: u16
}

impl Desktop {
    fn contains(&self, x: u16) -> bool {
        x >= self.position && x < self.position + self.width
    }
}

pub struct Bspwm {
    context: Context,
    tx: MessageSender,
    urgent_color: Color,
    selected_color: Color,
    desktops: Vec<Desktop>
}

pub fn bspwm(params: WidgetParams) -> Box<Widget> {
    let urgent_color = params.config.lookup("urgent_color")
        .and_then(|v| v.as_str())
        .and_then(|s| color::from_str(s))
        .unwrap_or(color::RED);
    let selected_color = params.config.lookup("selected_color")
        .and_then(|v| v.as_str())
        .and_then(|s| color::from_str(s))
        .unwrap_or(color::LIGHT_GREEN);
    let widget = Bspwm {
        context: params.context,
        tx: params.tx,
        desktops: vec![],
        urgent_color: urgent_color,
        selected_color: selected_color
    };
    Box::new(widget)
}

impl Bspwm {
    fn parse_bspwm(&mut self, line: &str) {
        let (kind, line) = line.split_at(1);
        if kind != "W" { return };

        let mut desktops = vec![];
        let elems = line.split(':');

        let mut pos = 0;

        for elem in elems {
            let mut chars = elem.chars();
            let kind = chars.next().unwrap();
            let name = chars.collect::<String>();

            if kind == 'M' || kind == 'm' {}
            else if kind == 'L' {}
            else if kind == 'G' {}
            else if kind == 'T' {}
            else {
                let empty = kind == 'f';
                let urgent = kind == 'U' || kind == 'u';
                let selected = kind.is_uppercase();

                if !empty {
                    let width = self.context.measure_text(&name) + MARGIN * 2;
                    desktops.push(Desktop {
                        name: name,
                        selected: selected,
                        urgent: urgent,
                        position: pos,
                        width: width
                    });
                    pos += width;
                }
            }
        }

        self.desktops = desktops;
    }
}

impl Widget for Bspwm {
    fn init(&mut self) {
        let tx = self.tx.clone();
        thread::spawn(move || monitor_thread(tx));
    }

    fn render(&mut self, x: u16, _w: u16) {
        for desktop in self.desktops.iter() {
            let color = if desktop.selected {
                self.selected_color
            }
            else if desktop.urgent {
                self.urgent_color
            }
            else {
                self.context.default_bg_color
            };
            self.context.set_bg_color(color);
            self.context.draw_bg(x + desktop.position, desktop.width);
            self.context.draw_text(&desktop.name, x + desktop.position + MARGIN);
        }
    }

    fn width(&mut self) -> u16 {
        let mut sum = 0;
        for desktop in self.desktops.iter() {
            sum += self.context.measure_text(&desktop.name) + MARGIN * 2;
        }
        sum
    }

    fn handle_event(&mut self, event: &Message) -> Update {
        match event {
            &Message::BspwmEvent(ref line) => {
                self.parse_bspwm(line);
                Update::Relayout
            },
            &Message::MousePress(x) => {
                for desktop in self.desktops.iter() {
                    if desktop.contains(x) {
                        switch_desktop(&desktop.name);
                    }
                }
                Update::Nothing
            },
            _ => Update::Nothing
        }
    }
}

fn monitor_thread(tx: MessageSender) {
    let bspc = Command::new("bspc")
        .arg("subscribe")
        .arg("report")
        .stdout(Stdio::piped())
        .spawn()
        .ok()
        .expect("Failed to start bspc subscribe");

    let stdout = bspc.stdout.unwrap();
    let mut reader = BufReader::new(stdout);

    let mut line = String::new();
    loop {
        line.clear();
        reader.read_line(&mut line).ok().expect("Failed to read line");
        let event = Message::BspwmEvent(line.clone());
        tx.send(event).ok();
    }
}

fn switch_desktop(name: &str) {
    Command::new("bspc")
        .arg("desktop")
        .arg("-f")
        .arg(name)
        .output()
        .ok();
}
